# -*- coding: utf-8 -*-
#
#    Copyright 2020 Ibai Roman
#
#    This file is part of EvoCov.
#
#    EvoCov is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    EvoCov is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with EvoCov. If not, see <http://www.gnu.org/licenses/>.

import numpy as np

import deap.gp
import deap.creator


def add_mutation(toolbox, pset, mutation_method='all'):
    """

    :param toolbox:
    :type toolbox:
    :param pset:
    :type pset:
    :param mutation_method:
    :type mutation_method:
    :return:
    :rtype:
    """

    toolbox.register(
        "mutate_uniform",
        deap.gp.mutUniform,
        expr=toolbox.random_mini_expr,
        pset=pset
    )
    toolbox.register(
        "mutate_replacement",
        deap.gp.mutNodeReplacement,
        pset=pset
    )
    toolbox.register(
        "mutate_shrink",
        deap.gp.mutShrink
    )
    toolbox.register(
        "mutate_insert",
        deap.gp.mutInsert,
        pset=pset
    )

    if mutation_method == "all":
        mutate_functions = [
            toolbox.mutate_uniform,
            toolbox.mutate_replacement,
            toolbox.mutate_shrink,
            # TODO toolbox.mutate_insert,
        ]
    else:
        raise Exception("This mutation method is not available")

    toolbox.register(
        "mutate",
        mut_kernel,
        mutate_functions=mutate_functions
    )
    toolbox.decorate("mutate", toolbox.remove_tuple_decorator)
    toolbox.decorate("mutate", toolbox.is_psd_decorator)
    toolbox.decorate("mutate", toolbox.inheritance_decorator)


def mut_kernel(old_individual, mutate_functions):
    """
    Replaces a randomly chosen primitive from *individual* by a randomly
    chosen primitive with the same number of arguments from the :attr:`pset`
    attribute of the individual.

    :param old_individual: The normal or typed tree to be mutated.
    :type old_individual:
    :param mutate_functions: mutate_functions:
    :type mutate_functions:
    :return: A tuple of one tree.
    :rtype:
    """

    mutate_function = np.random.choice(mutate_functions)

    old_prim_tree = deap.gp.PrimitiveTree(old_individual)

    new_prim_tree = mutate_function(old_prim_tree)[0]

    new_individual = deap.creator.Individual(new_prim_tree)

    new_individual.log.origin = "{} from {}".format(
        mutate_function.func.__name__,
        old_individual.log.id
    )

    if str(new_individual) == str(old_individual):
        return None,

    return new_individual,



from .random import Random
from .greedy import Greedy
from .evocov import EvoCov
from .moecov import MOECov
from .genetic_programming import GeneticProgramming
from .go_with_the_first import GoWithTheFirst


# -*- coding: utf-8 -*-
#
#    Copyright 2020 Ibai Roman
#
#    This file is part of EvoCov.
#
#    EvoCov is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    EvoCov is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with EvoCov. If not, see <http://www.gnu.org/licenses/>.

import numpy as np

import deap.algorithms

from .kernel_fitting_method import KernelFittingMethod
from ..gen_prog_tools import selection
from ..gen_prog_tools import mutation
from ..gen_prog_tools import crossover


class GeneticProgramming(KernelFittingMethod):
    """

    """

    def __init__(self, obj_fun, max_fun_call=500, dims=1,
                 nested_fit_method=None,
                 cxpb=0.4, mutpb=0.6, beta=1e-4):

        super(GeneticProgramming, self).__init__(
            obj_fun,
            max_fun_call=max_fun_call,
            dims=dims,
            nested_fit_method=nested_fit_method
        )

        selection.add_selection(
            toolbox=self.toolbox,
            selection_method='best'
        )

        mutation.add_mutation(
            toolbox=self.toolbox,
            pset=self.pset,
            mutation_method='all'
        )

        crossover.add_crossover(
            toolbox=self.toolbox,
            pset=self.pset
        )

        self.cxpb = cxpb
        self.mutpb = mutpb
        self.beta = beta

    def fit(self, model, folds, budget=None, verbose=False):
        """

        :param model:
        :type model:
        :param folds:
        :type folds:
        :param budget:
        :type budget:
        :param verbose:
        :type verbose:
        :return:
        :rtype:
        """

        max_eval, max_fun_call = self.get_max_eval(budget)
        ngen = int(np.power(max_eval, 0.535))
        npop = int(float(max_eval)/ngen)
        mu = max(3, int(npop * 0.25))

        best_state, initial_state = self.model_to_state(model)

        population = self.toolbox.random_population(n=npop)
        id_i = 0
        for ind in population:
            ind.log.id = id_i
            id_i += 1

        # Evaluate the individuals with an invalid fitness
        self.evaluate_population(
            model,
            best_state,
            initial_state,
            max_fun_call,
            folds,
            population,
            verbose=verbose
        )
        all_population = population

        # Begin the generational process
        for gen in range(0, ngen-1):
            # Select the next generation population
            selection = self.toolbox.select(population, mu)
            # Vary the population
            offspring = deap.algorithms.varOr(
                selection,
                self.toolbox,
                npop,
                self.cxpb,
                self.mutpb
            )
            for ind in offspring:
                ind.log.creation = gen + 1
                ind.log.id = id_i
                id_i += 1

            # Evaluate the individuals with an invalid fitness
            self.evaluate_population(
                model,
                best_state,
                initial_state,
                max_fun_call,
                folds,
                offspring,
                verbose=verbose
            )
            population = offspring + selection
            all_population += offspring

        best = self.toolbox.select(all_population, 1)[0]

        self.best_to_model(model, best_state, best)

        return best_state



# -*- coding: utf-8 -*-
#
#    Copyright 2020 Ibai Roman
#
#    This file is part of EvoCov.
#
#    EvoCov is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    EvoCov is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with EvoCov. If not, see <http://www.gnu.org/licenses/>.

from .kernel_fitting_method import KernelFittingMethod

from ..gen_prog_tools import selection
from ..gen_prog_tools import mutation


class GoWithTheFirst(KernelFittingMethod):
    """

    """

    def __init__(self, obj_fun, max_fun_call=500, dims=1,
                 nested_fit_method=None):

        super(GoWithTheFirst, self).__init__(
            obj_fun,
            max_fun_call=max_fun_call,
            dims=dims,
            nested_fit_method=nested_fit_method
        )

        selection.add_selection(
            toolbox=self.toolbox,
            selection_method='best'
        )

        mutation.add_mutation(
            toolbox=self.toolbox,
            pset=self.pset,
            mutation_method='all'
        )

    def fit(self, model, folds, budget=None, verbose=False):
        """

        :param model:
        :type model:
        :param folds:
        :type folds:
        :param budget:
        :type budget:
        :param verbose:
        :type verbose:
        :return:
        :rtype:
        """

        max_eval, max_fun_call = self.get_max_eval(budget)
        nmut = max(3, int(max_eval * 0.01))
        hof_size = 1
        npop = self.get_npop(
            hof_size,
            int(float(max_eval) / (nmut + 1))
        )

        best_state, initial_state = self.model_to_state(model)

        population = self.toolbox.random_population(n=npop)
        id_i = 0
        for ind in population:
            ind.log.id = id_i
            id_i += 1

        # Evaluate the individuals with an invalid fitness
        self.evaluate_population(
            model,
            best_state,
            initial_state,
            max_fun_call,
            folds,
            population,
            verbose=verbose
        )

        for gen_i in range(npop - hof_size + 1):
            best_pop = []
            pop_size = npop - gen_i
            pop_j = 0
            while pop_j < pop_size:
                best = population[pop_j]
                mut_k = 0
                while mut_k < nmut:
                    mutated_individual = self.toolbox.mutate(best)[0]
                    best.log.id = id_i
                    best.log.creation = gen_i + 1
                    id_i += 1
                    eval_population = [mutated_individual]
                    self.evaluate_population(
                        model,
                        best_state,
                        initial_state,
                        max_fun_call,
                        folds,
                        eval_population,
                        verbose=verbose
                    )
                    mutated_individual = eval_population[0]
                    best = self.toolbox.select([best, mutated_individual], 1)[0]
                    mut_k += 1
                best_pop.append(best)
                pop_j += 1

            population[:] = self.toolbox.select(
                best_pop,
                max(1, pop_size - 1)
            )

        best = population[0]

        self.best_to_model(model, best_state, best)

        return best_state

    @staticmethod
    def get_npop(hof_size, max_eval):
        """

        :param hof_size:
        :type hof_size:
        :param max_eval:
        :type max_eval:
        :return:
        :rtype:
        """
        n_pop = hof_size
        n_eval = hof_size
        while n_eval <= max_eval:
            n_pop += 1
            n_eval += n_pop

        return n_pop - 1



# -*- coding: utf-8 -*-
#
#    Copyright 2020 Ibai Roman
#
#    This file is part of EvoCov.
#
#    EvoCov is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    EvoCov is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with EvoCov. If not, see <http://www.gnu.org/licenses/>.

import collections

import numpy as np

import deap.gp


class InputX(object):
    pass


class OnlyDiagonal(object):
    pass


class HyperParameter(object):
    pass


# class Dimension(object):
#     pass


class ShapeX(object):
    pass


class TransX(object):
    pass


class CovMatrix(object):
    pass


def get_primitive_set(arg_num):
    """

    :param arg_num:
    :type arg_num:
    :return:
    :rtype:
    """
    pset = deap.gp.PrimitiveSetTyped(
        "MAIN",
        [InputX, OnlyDiagonal] + ([HyperParameter] * arg_num),
        CovMatrix
    )
    pset.renameArguments(ARG0='x')
    pset.renameArguments(ARG1='onlyvar')
    for i in range(arg_num):
        pset.renameArguments(**{'ARG{}'.format(i+2): 'hp{}'.format(i)})
    # for i in range(20):
    #     pset.addEphemeralConstant(
    #         'ec{}'.format(i),
    #         lambda: np.exp(np.random.uniform(-10, 10)),
    #         HyperParameter
    #     )

    #  TERMNINALS  #

    # HyperParameter Type
    pset.addTerminal(0.5, HyperParameter)
    pset.addTerminal(1.0, HyperParameter)
    pset.addTerminal(2.0, HyperParameter)
    pset.addTerminal(3.0, HyperParameter)
    pset.addTerminal(5.0, HyperParameter)

    # Dimension Type
    # for i in range(dims):
    #     pset.addTerminal(i, Dimension)

    #  PRIMITIVES  #

    # ShapeX Type
    pset.addPrimitive(
        x_shape,
        [InputX],
        ShapeX,
        "x_shape"
    )

    # TransX Type
    pset.addPrimitive(
        lambda x: x,
        [InputX],
        TransX,
        "map"
    )
    # pset.addPrimitive(
    #     x_mask,
    #     [InputX, Dimension],
    #     TransX,
    #     "x_mask"
    # )
    pset.addPrimitive(
        x_div,
        [TransX, HyperParameter],
        TransX,
        "x_div"
    )
    pset.addPrimitive(
        x_sub,
        [TransX, HyperParameter],
        TransX,
        "x_sub"
    )
    pset.addPrimitive(
        spectral,
        [TransX],
        TransX,
        "spectral"
    )

    # CovMatrix Type
    pset.addPrimitive(
        sq_distance,
        [TransX, OnlyDiagonal],
        CovMatrix,
        "sq_distance"
    )
    pset.addPrimitive(
        dot_product,
        [TransX, OnlyDiagonal],
        CovMatrix,
        "dot_product"
    )
    pset.addPrimitive(
        constant,
        [ShapeX, HyperParameter, OnlyDiagonal],
        CovMatrix,
        "constant"
    )
    pset.addPrimitive(
        noise,
        [ShapeX, HyperParameter, OnlyDiagonal],
        CovMatrix,
        "noise"
    )
    # pset.addPrimitive(
    #     polynomial,
    #     [CovMatrix, HyperParameter, Dimension],
    #     CovMatrix,
    #     "polynomial"
    # )
    # pset.addPrimitive(
    #     comp_fun,
    #     [TransX, OnlyDiagonal],
    #     CovMatrix,
    #     "comp_fun"
    # )
    # pset.addPrimitive(
    #     change_window,
    #     [
    #         MaskedX,
    #         CovMatrix, CovMatrix,
    #         HyperParameter, HyperParameter, HyperParameter, OnlyDiagonal
    #     ],
    #     CovMatrix,
    #     "cw"
    # )
    # pset.addPrimitive(
    #     change_point,
    #     [
    #         MaskedX,
    #         CovMatrix, CovMatrix,
    #         HyperParameter, HyperParameter, OnlyDiagonal
    #     ],
    #     CovMatrix,
    #     "cp"
    # )
    pset.addPrimitive(np.add, [CovMatrix, CovMatrix], CovMatrix)
    pset.addPrimitive(np.multiply, [CovMatrix, CovMatrix], CovMatrix)
    pset.addPrimitive(np.exp, [CovMatrix], CovMatrix)
    pset.addPrimitive(np.negative, [CovMatrix], CovMatrix)
    pset.addPrimitive(np.tanh, [CovMatrix], CovMatrix)
    pset.addPrimitive(np.power, [CovMatrix, HyperParameter], CovMatrix)
    pset.addPrimitive(
        lambda x: np.power(x, -1.0),
        [CovMatrix],
        CovMatrix,
        "div"
    )
    pset.addPrimitive(np.sqrt, [CovMatrix], CovMatrix)
    pset.addPrimitive(np.square, [CovMatrix], CovMatrix)

    pset.cont_primitives = collections.defaultdict(list)
    pset.end_primitives = collections.defaultdict(list)
    for type_, primitives in pset.primitives.items():
        pset.cont_primitives[type_] = []
        pset.end_primitives[type_] = []
        for prim in primitives:
            if np.any([arg == type_ for arg in prim.args]):
                pset.cont_primitives[type_].append(prim)
            else:
                pset.end_primitives[type_].append(prim)

    return pset


def x_shape(x):
    """

    :param x:
    :type x:
    :return:
    :rtype:
    """

    mat_a, mat_b = x

    len_a = len(mat_a)
    if mat_b is None:
        len_b = None
    else:
        len_b = len(mat_b)

    return len_a, len_b


def x_div(x, ls):
    """

    :param x:
    :type x:
    :param ls:
    :type ls:
    :return:
    :rtype:
    """
    mat_a, mat_b = x

    t_mat_a = mat_a / ls
    if mat_b is not None:
        t_mat_b = mat_b / ls
    else:
        t_mat_b = None

    return t_mat_a, t_mat_b


def x_sub(x, shift):
    """

    :param x:
    :type x:
    :param shift:
    :type shift:
    :return:
    :rtype:
    """
    mat_a, mat_b = x

    t_mat_a = mat_a - shift
    if mat_b is not None:
        t_mat_b = mat_b - shift
    else:
        t_mat_b = None

    return t_mat_a, t_mat_b


# def x_mask(x, dim):
#     """
#
#     :param x:
#     :type x:
#     :param dim:
#     :type dim:
#     :return:
#     :rtype:
#     """
#     mat_a, mat_b = x
#     total_dims = mat_a.shape[1]
#     masked_dim = int(dim)
#     mask = np.full(total_dims, np.nan)
#     mask[masked_dim] = 1.0
#     t_mat_a = mat_a * mask
#     if mat_b is not None:
#         t_mat_b = mat_b * mask
#     else:
#         t_mat_b = None
#
#     return t_mat_a, t_mat_b


def spectral(transformed_x):
    """

    :param transformed_x:
    :type transformed_x: tuple
    :return:
    :rtype: tuple
    """
    mat_a, mat_b = transformed_x

    a_param = 2.0 * np.pi

    spectral_mat_a = np.hstack((
        np.sin(a_param * mat_a),
        np.cos(a_param * mat_a)
    ))

    spectral_mat_b = None
    if mat_b is not None:
        spectral_mat_b = np.hstack((
            np.sin(a_param * mat_b),
            np.cos(a_param * mat_b)
        ))

    spectral_x = (spectral_mat_a, spectral_mat_b)

    return spectral_x


def sq_distance(transformed_x, only_diagonal):
    """

    :param transformed_x:
    :type transformed_x: tuple
    :param only_diagonal:
    :type only_diagonal:
    :return:
    :rtype: np.ndarray
    """
    mat_a, mat_b = transformed_x

    mat_a = np.nan_to_num(mat_a)
    if mat_b is None:
        if only_diagonal:
            return np.zeros((len(mat_a), 1))

        center = np.mean(mat_a, axis=0)
        mat_a = mat_a - center
        mat_a_sq = np.sum(np.square(mat_a), 1)

        r2 = mat_a_sq[:, None] + mat_a_sq[None, :]

        r2 -= 2. * np.dot(mat_a, mat_a.T)

        np.lib.stride_tricks.as_strided(
            r2, shape=(r2.shape[0], ),
            strides=((r2.shape[0] + 1) * r2.itemsize, )
        )[:, ] = 0
    else:
        mat_b = np.nan_to_num(mat_b)
        center = np.mean(np.vstack((mat_a, mat_b)), axis=0)
        mat_a = mat_a - center
        mat_b = mat_b - center
        mat_a_sq = np.sum(np.square(mat_a), 1)
        mat_b_sq = np.sum(np.square(mat_b), 1)

        r2 = mat_a_sq[:, None] + mat_b_sq[None, :]

        r2 -= 2. * np.dot(mat_a, mat_b.T)

    r2 = r2.clip(min=0.0)
    return r2


def dot_product(transformed_x, only_diagonal):
    """

    :param transformed_x:
    :type transformed_x: tuple
    :param only_diagonal:
    :type only_diagonal: bool
    :return:
    :rtype: np.ndarray
    """
    mat_a, mat_b = transformed_x

    mat_a = np.nan_to_num(mat_a)
    if mat_b is None:
        if only_diagonal:
            return np.sum(np.square(mat_a), axis=1)[:, None]
        mat_b = mat_a
    else:
        mat_b = np.nan_to_num(mat_b)
    return np.dot(mat_a, mat_b.T)


def constant(shape_x, ov, only_diagonal):
    """

    :param shape_x:
    :type shape_x: tuple
    :param ov:
    :type ov: float
    :param only_diagonal:
    :type only_diagonal:
    :return:
    :rtype: np.ndarray
    """

    len_a, len_b = shape_x

    if len_b is None:
        len_b = len_a
        if only_diagonal:
            len_b = 1

    return ov * np.ones((len_a, len_b))


def noise(shape_x, ov, only_diagonal):
    """

    :param shape_x:
    :type shape_x: tuple
    :param ov:
    :type ov: float
    :param only_diagonal:
    :type only_diagonal:
    :return:
    :rtype: np.ndarray
    """

    len_a, len_b = shape_x

    if len_b is not None:
        return np.zeros((len_a, len_b))

    if only_diagonal:
        return ov * np.ones((len_a, 1))

    return ov * np.eye(len_a)


# def polynomial(cov, coef, dim):
#     """
#
#     :param cov:
#     :type cov:
#     :param coef:
#     :type coef:
#     :param dim:
#     :type dim:
#     :return:
#     :rtype:
#     """
#
#     return np.multiply(coef, np.power(cov, dim))
#
#
# def comp_fun(transformed_x, only_diagonal):
#     """
#
#     :param transformed_x:
#     :type transformed_x: tuple
#     :param only_diagonal:
#     :type only_diagonal:
#     :return:
#     :rtype: np.ndarray
#     """
#     mat_a, mat_b = transformed_x
#
#     if mat_b is None:
#         if only_diagonal:
#             return np.ones((len(mat_a), 1))
#
#         center = np.mean(mat_a, axis=0)
#         mat_a = mat_a - center
#         mat_a_sq = -0.5 * np.sum(np.square(mat_a), 1)
#
#         r2 = mat_a_sq[:, None] + mat_a_sq[None, :]
#
#         r2 += np.dot(mat_a, mat_a.T)
#
#     else:
#         center = np.mean(np.vstack((mat_a, mat_b)), axis=0)
#         mat_a = mat_a - center
#         mat_b = mat_b - center
#         mat_a_sq = -0.5 * np.sum(np.square(mat_a), 1)
#         mat_b_sq = -0.5 * np.sum(np.square(mat_b), 1)
#
#         r2 = mat_a_sq[:, None] + mat_b_sq[None, :]
#
#         r2 += np.dot(mat_a, mat_b.T)
#
#     return np.exp(r2)
#
#
# def change_window(masked_x, cov1, cov2,
#                   location, steepness, width, only_diagonal):
#     """
#
#     :param masked_x:
#     :type masked_x:
#     :param cov1:
#     :type cov1:
#     :param cov2:
#     :type cov2:
#     :param location:
#     :type location:
#     :param steepness:
#     :type steepness:
#     :param width:
#     :type width:
#     :param only_diagonal:
#     :type only_diagonal:
#     :return:
#     :rtype:
#     """
#     mat_a, mat_b = masked_x
#
#     tx1 = np.tanh(
#         (mat_a - (location - 0.5 * width)) * steepness
#     )
#     tx2 = np.tanh(
#         -(mat_a - (location + 0.5 * width)) * steepness
#     )
#     ax = np.multiply((0.5 + 0.5 * tx1), (0.5 + 0.5 * tx2))
#     if mat_b is not None:
#         tz1 = np.tanh(
#             (mat_b - (location - 0.5 * width)) * steepness
#         )
#         tz2 = np.tanh(
#             -(mat_b - (location + 0.5 * width)) * steepness
#         )
#         az = np.multiply((0.5 + 0.5 * tz1), (0.5 + 0.5 * tz2)).T
#     else:
#         if only_diagonal:
#             az = ax
#         else:
#             az = ax.T
#
#     k_matrix = np.multiply(np.multiply(ax, cov2), az) + \
#         np.multiply(np.multiply((1 - ax), cov1), (1 - az))
#
#     return k_matrix
#
#
# def change_point(masked_x, cov1, cov2, location, steepness, only_diagonal):
#     """
#
#     :param masked_x:
#     :type masked_x:
#     :param cov1:
#     :type cov1:
#     :param cov2:
#     :type cov2:
#     :param location:
#     :type location:
#     :param steepness:
#     :type steepness:
#     :param only_diagonal:
#     :type only_diagonal:
#     :return:
#     :rtype:
#     """
#     mat_a, mat_b = masked_x
#
#     tx = np.tanh(
#         (mat_a - location) * steepness
#     )
#     ax = 0.5 + 0.5 * tx
#     if mat_b is not None:
#         tz = np.tanh(
#             (mat_b - location) * steepness
#         )
#         az = (0.5 + 0.5 * tz).T
#     else:
#         if only_diagonal:
#             az = ax
#         else:
#             az = ax.T
#
#     ax = 1 - ax
#     az = 1 - az
#
#     k_matrix = np.multiply(np.multiply(ax, cov1), az) + \
#         np.multiply(np.multiply((1 - ax), cov2), (1 - az))
#
#     return k_matrix



# -*- coding: utf-8 -*-
#
#   Copyright 2020 Ibai Roman
#
#   This file is part of EvoCov.
#
#   EvoCov is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   EvoCov is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with EvoCov. If not, see <http://www.gnu.org/licenses/>.

import numpy as np

import deap.gp
import deap.creator

from .primitive_set import CovMatrix


def add_crossover(toolbox, pset):
    """

    :param toolbox:
    :type toolbox:
    :param pset:
    :type pset:
    :return:
    :rtype:
    """
    toolbox.register(
        "mate",
        cx_kernel,
        pset=pset
    )
    toolbox.decorate("mate", toolbox.remove_tuple_decorator)
    toolbox.decorate("mate", toolbox.is_psd_decorator)
    toolbox.decorate("mate", toolbox.inheritance_decorator)


def cx_kernel(old_ind1, old_ind2, pset):
    """
    Randomly select in each individual and exchange each subtree with the
    point as root between each individual.

    :param old_ind1: First tree participating in the crossover.
    :type old_ind1:
    :param old_ind2: Second tree participating in the crossover.
    :type old_ind2:
    :param pset:
    :type pset:
    :returns: A tuple of two trees.
    :rtype:
    """

    pri_name = np.random.choice(['multiply', 'add'])

    old_prim_tree1 = deap.gp.PrimitiveTree(old_ind1)
    old_prim_tree2 = deap.gp.PrimitiveTree(old_ind2)

    old_prim_tree1 = get_typed_subtree(old_prim_tree1, CovMatrix)
    old_prim_tree2 = get_typed_subtree(old_prim_tree2, CovMatrix)

    new_prim_tree = merge_inds(pri_name, old_prim_tree1, old_prim_tree2, pset)

    new_individual = deap.creator.Individual(new_prim_tree)

    new_individual.log.origin = "cx_kernel from {} and {}".format(
        old_ind1.log.id,
        old_ind2.log.id
    )

    return new_individual,


def get_typed_subtree(individual, _type):
    """

    :param individual:
    :type individual:
    :param _type:
    :type _type:
    :return:
    :rtype:
    """
    if np.random.choice([False, True]):
        return individual

    subtree = []
    return_type = object
    while return_type != _type:
        subtree = individual[
            individual.searchSubtree(np.random.randint(0, len(individual) - 1))
        ]
        return_type = subtree[0].ret

    individual[slice(0, None, None)] = subtree[slice(0, None, None)]

    return individual


def merge_inds(pri_name, ind1, ind2, pset):
    """

    :param pri_name:
    :type pri_name:
    :param ind1:
    :type ind1:
    :param ind2:
    :type ind2:
    :param pset:
    :type pset:
    :return:
    :rtype:
    """
    primitive = select_element(
        pset, element_type="primitive", _type=CovMatrix, name=pri_name
    )[0]
    ind12 = ind1[:] + ind2[:]
    new_expr = [primitive]
    new_expr += ind12

    ind1[slice(0, None, None)] = new_expr[slice(0, None, None)]

    return ind1


def select_element(pset, element_type=None, _type=None, name=None):
    """

    :param pset:
    :type pset:
    :param element_type:
    :type element_type:
    :param _type:
    :type _type:
    :param name:
    :type name:
    :return:
    :rtype:
    """

    if element_type == "terminal":
        elements = pset.terminals
    elif element_type == "primitive":
        elements = pset.primitives
    else:
        raise Exception("element not found")

    if type is not None:
        elements = elements[_type]

    if name is not None:
        elements = [
            element
            for element in elements
            if element.name == name
        ]

    return [np.random.choice(elements)]



# -*- coding: utf-8 -*-
#
#    Copyright 2020 Ibai Roman
#
#    This file is part of EvoCov.
#
#    EvoCov is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    EvoCov is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with EvoCov. If not, see <http://www.gnu.org/licenses/>.

from .kernel_fitting_method import KernelFittingMethod

from ..gen_prog_tools import selection
from ..gen_prog_tools import mutation


class Greedy(KernelFittingMethod):
    """

    """

    def __init__(self, obj_fun, max_fun_call=500, dims=1,
                 nested_fit_method=None):

        super(Greedy, self).__init__(
            obj_fun,
            max_fun_call=max_fun_call,
            dims=dims,
            nested_fit_method=nested_fit_method
        )

        selection.add_selection(
            toolbox=self.toolbox,
            selection_method='best'
        )

        mutation.add_mutation(
            toolbox=self.toolbox,
            pset=self.pset,
            mutation_method='all'
        )

    def fit(self, model, folds, budget=None, verbose=False):
        """

        :param model:
        :type model:
        :param folds:
        :type folds:
        :param budget:
        :type budget:
        :param verbose:
        :type verbose:
        :return:
        :rtype:
        """

        _, max_fun_call = self.get_max_eval(budget)

        best_state, initial_state = self.model_to_state(model)
        best_state['restarts'] = 0
        individual = model.get_kernel_function()

        base_id = int(individual.log.id * 1e5)
        best = individual
        population = []
        while best_state['fun_calls'] < max_fun_call:
            individual.log.id = base_id + best_state['restarts']
            # run optimization
            eval_population = [individual]
            self.evaluate_population(
                model,
                best_state,
                initial_state,
                max_fun_call,
                folds,
                eval_population,
                verbose=verbose
            )
            individual = eval_population[0]
            best_state['restarts'] += 1

            population.append(individual)
            best = self.toolbox.select(population, 1)[0]
            population = [best]
            individual = self.toolbox.mutate(best)[0]

        self.best_to_model(model, best_state, best)

        return best_state



# -*- coding: utf-8 -*-
#
#    Copyright 2020 Ibai Roman
#
#    This file is part of EvoCov.
#
#    EvoCov is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    EvoCov is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with EvoCov. If not, see <http://www.gnu.org/licenses/>.

import deap.gp

from gplib.parameters.parameter import Parameter
from gplib.parameters.log_parameter_transformation \
    import LogParameterTransformation
from gplib.kernel_functions.kernel_function import KernelFunction


class GenProgKernel(deap.gp.PrimitiveTree, KernelFunction):
    """

    """

    def __init__(self, content):
        deap.gp.PrimitiveTree.__init__(self, content)

        self.complied_function = deap.gp.compile(self, self.pset)
        individual_args = [
            item.value
            for item in self
            if item.arity == 0 and item.conv_fct == str
        ]
        has_hps = [
            hp if hp in individual_args else None
            for hp in self.pset.arguments
            if 'hp' in hp
        ]

        hyperparams = []
        hp_arg = []

        for hp in has_hps:
            if hp:
                hyperparam = Parameter(
                    name=hp,
                    transformation=LogParameterTransformation,
                    default_value=1.0
                )
                hyperparams.append(hyperparam)
            else:
                hyperparam = None

            hp_arg.append(hyperparam)
        self.hp_arg = hp_arg

        KernelFunction.__init__(self, hyperparams)

        self.log.prim_count = sum([
            1
            for item in self
            if isinstance(item, deap.gp.Primitive)
        ])

        self.log.hp_count = self.get_param_n()

    def __hash__(self):
        hp_hashes = [hash(self.complied_function)]
        for hp in self.hyperparameters:
            if isinstance(hp, Parameter):
                if hp.is_array():
                    hp_hashes.extend(hash(tuple(hp.current_value)))
                else:
                    hp_hashes.append(hash(hp.current_value))
            else:
                raise Exception("Incorrect covariance structure")
        return hash(tuple(hp_hashes))

    def function(self, mat_a, mat_b=None, only_diagonal=False):
        """
        Measures the distance matrix between solutions of A and B, and
        applies the kernel function element-wise to the distance matrix.

        :param mat_a: List of solutions in lines and dimensions in columns.
        :type mat_a:
        :param mat_b: List of solutions in lines and dimensions in columns.
        :type mat_b:
        :param only_diagonal:
        :type only_diagonal:
        :return: Result matrix with kernel function applied element-wise.
        :rtype:
        """
        variables = [(mat_a, mat_b), only_diagonal] + [
            hp.get_param_values()[0] if hp else 0.0 for hp in self.hp_arg]

        return self.complied_function(*variables)

    def dk_dx(self, mat_a, mat_b=None):
        """
        Measures gradient of the distance between solutions of A and B in X.

        :param mat_a: List of solutions in lines and dimensions in columns.
        :type mat_a:
        :param mat_b: List of solutions in lines and dimensions in columns.
        :type mat_b:
        :return: 3D array with the gradient in every dimension of X.
        :rtype:
        """

        raise NotImplementedError("Not Implemented. This is an interface.")

    def dk_dtheta(self, mat_a, mat_b=None, trans=False):
        """
        Measures gradient of the distance between solutions of A and B in the
        hyper-parameter space.

        :param mat_a: List of solutions in lines and dimensions in columns.
        :type mat_a:
        :param mat_b: List of solutions in lines and dimensions in columns.
        :type mat_b:
        :param trans: Return results in the transformed space.
        :type trans:
        :return: 3D array with the gradient in every
         dimension the length-scale hyper-parameter space.
        :rtype:
        """

        raise NotImplementedError("Not Implemented. This is an interface.")



# -*- coding: utf-8 -*-
#
#    Copyright 2020 Ibai Roman
#
#    This file is part of EvoCov.
#
#    EvoCov is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    EvoCov is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with EvoCov. If not, see <http://www.gnu.org/licenses/>.

import operator
import random
from inspect import isclass
from functools import wraps

import deap.tools
import deap.base
import deap.creator
import deap.gp

import numpy as np

import gplib

from .primitive_set import CovMatrix, HyperParameter
from ..kernel_functions.gen_prog_kernel import GenProgKernel


def add_creation(toolbox, pset, max_depth, dims):
    """
    
    :param toolbox: 
    :type toolbox: 
    :param pset: 
    :type pset: 
    :param max_depth: 
    :type max_depth: 
    :param dims: 
    :type dims: 
    :return: 
    :rtype: 
    """

    deap.creator.create(
        "FitnessMin",
        deap.base.Fitness,
        weights=tuple(-1.0 for _ in range(20))
    )

    deap.creator.create(
        "Log",
        object,
        id=0,
        evals=0,
        creation=0,
        hp_count=0,
        prim_count=0,
        origin=""
    )

    deap.creator.create(
        "Individual",
        GenProgKernel,
        pset=pset,
        fitness=deap.creator.FitnessMin,
        log=deap.creator.Log,
        nested=None
    )

    toolbox.register(
        "is_psd_decorator",
        is_psd_decorator,
        key=operator.attrgetter("height"),
        max_value=max_depth,
        dims=dims,
        check_noise_con=True,
        verbose=False
    )

    toolbox.register(
        "is_noise_con_psd_cov_decorator",
        is_psd_decorator,
        key=operator.attrgetter("height"),
        max_value=max_depth,
        dims=dims,
        check_noise_con=False,
        verbose=False
    )

    toolbox.register(
        "inheritance_decorator",
        inheritance_decorator,
        pset=pset
    )

    toolbox.register(
        "remove_tuple_decorator",
        remove_tuple_decorator
    )

    toolbox.register(
        "individual_from_string",
        deap.gp.PrimitiveTree.from_string,
        pset=pset
    )
    toolbox.decorate("individual_from_string", init_ind_decorator)
    toolbox.decorate(
        "individual_from_string",
        toolbox.is_noise_con_psd_cov_decorator
    )
    toolbox.decorate("individual_from_string", hps_decorator)

    toolbox.register(
        "random_expr",
        gen_typed,
        pset=pset,
        min_=3,
        max_=15
    )
    toolbox.register(
        "random_mini_expr",
        gen_typed,
        pset=pset,
        min_=3,
        max_=6
    )
    toolbox.register(
        "random_individual",
        toolbox.random_expr
    )
    toolbox.decorate("random_individual", init_ind_decorator)
    toolbox.decorate(
        "random_individual",
        toolbox.is_psd_decorator
    )

    toolbox.register(
        "random_population",
        deap.tools.initRepeat,
        list,
        toolbox.random_individual
    )


def is_psd_decorator(func, key, max_value, dims,
                     check_noise_con=True, verbose=True):
    """

    :param func:
    :type func:
    :param key:
    :type key:
    :param max_value:
    :type max_value:
    :param dims:
    :type dims:
    :param check_noise_con:
    :type check_noise_con:
    :param verbose:
    :type verbose:
    :return:
    :rtype:
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        trial = 0
        while trial < 600:
            trial += 1
            candidate_individual = func(*args, **kwargs)
            if not candidate_individual:
                continue
            candidate_individual = is_psd(
                candidate_individual,
                dims,
                check_noise_con=check_noise_con,
                verbose=verbose
            )
            if not candidate_individual:
                continue
            if key(candidate_individual) <= max_value:
                return candidate_individual
        return None

    return wrapper


def is_psd(individual, dims, items=200, n_data_sets=50,
           check_noise_con=True, verbose=True):
    """

    :param individual:
    :type individual:
    :param dims:
    :type dims:
    :param items:
    :type items:
    :param n_data_sets:
    :type n_data_sets:
    :param check_noise_con:
    :type check_noise_con:
    :param verbose:
    :type verbose:
    :return:
    :rtype:
    """

    if verbose:
        print("s) {0}".format(str(individual)))

    saved_params = individual.get_param_values(trans=False)
    n_hp_trials = (10 * individual.get_param_n()) + 1

    i_hp_trials = 0
    individual_is_psd = False
    def_opt_values = individual.get_param_values(
        only_group=gplib.hp.parameter.Parameter.OPT_GROUP,
        trans=True
    )
    def_grid_values = individual.get_grid(
        only_group=gplib.hp.parameter.Parameter.GRID_GROUP,
        trans=False
    )
    while not individual_is_psd and i_hp_trials < n_hp_trials:
        if 0 < len(def_grid_values):
            individual.set_param_values(
                random.choice(def_grid_values),
                only_group=gplib.hp.parameter.Parameter.GRID_GROUP,
                trans=False
            )
        current_value = def_opt_values + np.array(
            np.random.normal(
                loc=0.0,
                scale=1.,
                size=len(def_opt_values)
            )
        )
        individual.set_param_values(
            current_value,
            only_group=gplib.hp.parameter.Parameter.OPT_GROUP,
            trans=True
        )
        individual_is_psd = True
        i_data_set = 0
        while individual_is_psd and i_data_set < n_data_sets:
            data = np.vstack((
                np.zeros(dims),
                np.random.rand(1, dims) +
                np.random.randn(int(items / 2.0), dims) * 1e-3,
                np.random.rand(int(items / 2.0), dims)
            ))
            cov_matrix = individual.function(data)
            try:
                if check_noise_con:
                    is_noise_con(cov_matrix)
                gplib.MGD._chol(cov_matrix)
            except np.linalg.LinAlgError:
                individual_is_psd = False
            i_data_set += 1
        i_hp_trials += 1

    individual.set_param_values(saved_params, trans=False)

    if not individual_is_psd:
        if verbose:
            print("\t[FAIL] Is not PSD")
        return None

    if verbose:
        print("\t[OK]")

    return individual


def is_noise_con(cov_matrix):
    """

    :param cov_matrix:
    :type cov_matrix:
    :return:
    :rtype:
    """

    # Cov matrix is Noise + Constant
    diag_index = np.eye(cov_matrix.shape[0], dtype=bool)
    diag_matrix = cov_matrix[diag_index]
    diagless_matrix = cov_matrix[np.logical_not(diag_index)]
    diag_diff = np.max(diag_matrix) - np.min(diag_matrix)
    diagless_diff = np.max(diagless_matrix) - np.min(diagless_matrix)
    if diag_diff < 1e-20 and diagless_diff < 1e-20:
        raise np.linalg.LinAlgError("Cov matrix is Noise + Constant")


def inheritance_decorator(func, pset):
    """

    :param func:
    :type func:
    :param pset:
    :type pset:
    :return:
    :rtype:
    """
    @wraps(func)
    def wrapper(*args, **kwargs):
        old_inds = [
            ind
            for ind in args
            if isinstance(ind, deap.creator.Individual)
        ]
        if len(old_inds) == 2:
            change_ind2(old_inds[0], old_inds[1], pset)
        hps_to_inherit = {
            cov_hp.get_param_keys(recursive=False): cov_hp.get_param_values()[0]
            for old_ind in old_inds
            for cov_hp in old_ind.get_hyperparams()
        }

        candidate_individual = func(*args, **kwargs)

        new_inds = []
        if candidate_individual:
            load_hyperparams_to_cov(
                candidate_individual,
                hps_to_inherit
            )
            new_inds.append(candidate_individual)

        new_inds += old_inds[len(new_inds):]
        return tuple(new_inds)

    return wrapper


def change_ind2(ind1, ind2, pset):
    """

    :param ind1:
    :type ind1:
    :param ind2:
    :type ind2:
    :param pset:
    :type pset:
    :return:
    :rtype:
    """
    hps_in_ind1 = set(
        terminal.value for terminal in ind1[:]
        if type(terminal) == deap.gp.Terminal and 'hp' in str(terminal.value)
    )
    hps_in_ind2 = set(
        terminal.value for terminal in ind2[:]
        if type(terminal) == deap.gp.Terminal and 'hp' in str(terminal.value)
    )
    avaiable_hps = list(
        terminal.value
        for terminal in pset.terminals[HyperParameter]
        if terminal.value not in hps_in_ind1 and
        terminal.value not in hps_in_ind2 and 'hp' in str(terminal.value)
    )
    avaiable_hps.sort()
    random.shuffle(avaiable_hps)
    hps_to_remove = list(hps_in_ind1.intersection(hps_in_ind2))
    hps_to_remove.sort()
    random.shuffle(hps_to_remove)

    # Replace from ind2 the parameters that are also in ind1
    # Add them also to the new parameter dict
    for hp_to_remove in hps_to_remove:
        if avaiable_hps:
            hp_to_put = avaiable_hps.pop()
        else:
            hp_to_put = hp_to_remove
            print("No HPS left")
        for i in range(len(ind2[:])):
            if type(ind2[i]) == deap.gp.Terminal and \
                    ind2[i].value == hp_to_remove:
                ind2[i] = [
                    terminal
                    for terminal in pset.terminals[HyperParameter]
                    if hp_to_put == str(terminal.value)
                ][0]
        for hyperparam in ind2.get_hyperparams():
            if hyperparam.name == hp_to_remove:
                hyperparam.name = hp_to_put


def remove_tuple_decorator(func):
    """

    :param func:
    :type func:
    :return:
    :rtype:
    """
    @wraps(func)
    def wrapper(*args, **kwargs):
        ret_tuple = func(*args, **kwargs)

        individual = ret_tuple[0]

        return individual
    return wrapper


def init_ind_decorator(func):
    """

    :param func:
    :type func:
    :return:
    :rtype:
    """
    @wraps(func)
    def wrapper(*args, **kwargs):
        expr = func(*args, **kwargs)
        individual = deap.creator.Individual(expr)
        individual.log.origin = func.__name__
        return individual
    return wrapper


def hps_decorator(func):
    """

    :param func:
    :type func:
    :return:
    :rtype:
    """
    @wraps(func)
    def wrapper(*args, **kwargs):
        hps = {}
        if 'hps' in kwargs:
            hps.update(kwargs['hps'])
            del kwargs['hps']

        individual = func(*args, **kwargs)

        load_hyperparams_to_cov(individual, hps)

        return individual
    return wrapper


def load_hyperparams_to_cov(individual, hps_to_inherit):
    """

    :param individual:
    :type individual:
    :param hps_to_inherit:
    :type hps_to_inherit:
    :return:
    :rtype:
    """

    if hps_to_inherit:
        for cov_hp in individual.get_hyperparams():
            name = cov_hp.get_param_keys(recursive=False)
            if name in hps_to_inherit:
                cov_hp.set_param_values(
                    [hps_to_inherit[name]]
                )


def gen_typed(pset, min_, max_, type_=None):
    """

    :param pset:
    :type pset:
    :param min_:
    :type min_:
    :param max_:
    :type max_:
    :param type_:
    :type type_:
    :return:
    :rtype:
    """
    if type_ is None:
        type_ = CovMatrix

    options = []
    desirable_options = []
    if len(pset.terminals[type_]) > 0:
        options.append('add_terminal')
        if 2 >= min_:
            desirable_options.append('add_terminal')
    if len(pset.cont_primitives[type_]) > 0:
        options.append('add_cont_primitive')
        if max_ >= 2:
            desirable_options.append('add_cont_primitive')
    if len(pset.end_primitives[type_]) > 0:
        options.append('add_end_primitive')
        if 2 >= min_:
            desirable_options.append('add_end_primitive')

    if len(options) == 1:
        result = options[0]
    elif len(options) < 1:
        raise Exception("No options left")
    else:
        if len(desirable_options) == 1:
            result = desirable_options[0]
        elif len(desirable_options) < 1:
            result = np.random.choice(options)
        else:
            result = np.random.choice(desirable_options)

    expr = []
    if result == 'add_terminal':
        term = np.random.choice(
            pset.terminals[type_]
        )

        if isclass(term):
            term = term()
        expr.append(term)
    if result in ['add_cont_primitive', 'add_end_primitive']:
        if result == 'add_cont_primitive':
            prim = np.random.choice(
                pset.cont_primitives[type_]
            )
        else:
            prim = np.random.choice(
                pset.end_primitives[type_]
            )
        expr.append(prim)
        for arg_type in prim.args:
            arg = gen_typed(
                pset, min_ - 1, max_ - 1, type_=arg_type
            )
            expr.extend(arg)

    return expr



# -*- coding: utf-8 -*-
#
#    Copyright 2020 Ibai Roman
#
#    This file is part of EvoCov.
#
#    EvoCov is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    EvoCov is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with EvoCov. If not, see <http://www.gnu.org/licenses/>.

import time

import numpy as np

import deap.base
import deap.creator

from gplib.fitting_methods.fitting_method import FittingMethod
from ..gen_prog_tools import primitive_set, creation


class KernelFittingMethod(FittingMethod):
    """

    """
    MAX_SIGMA = 10.0 ** 20

    def __init__(self, obj_fun, max_fun_call=500, dims=1,
                 nested_fit_method=None):

        self.max_fun_call = max_fun_call

        self.obj_fun = obj_fun

        self.toolbox = deap.base.Toolbox()

        self.pset = primitive_set.get_primitive_set(
            arg_num=20
        )

        creation.add_creation(
            toolbox=self.toolbox,
            pset=self.pset,
            max_depth=30,
            dims=dims
        )

        super(KernelFittingMethod, self).__init__(
            nested_fit_method=nested_fit_method
        )

    def evaluate_population(self, model, best_state, initial_state,
                            max_fun_call, folds,
                            population, all_population=None, verbose=False):
        """

        :param model:
        :type model:
        :param best_state:
        :type best_state:
        :param initial_state:
        :type initial_state:
        :param max_fun_call:
        :type max_fun_call:
        :param folds:
        :type folds:
        :param population:
        :type population:
        :param all_population:
        :type all_population:
        :param verbose:
        :type verbose:
        :return:
        :rtype:
        """
        for pop_i in range(len(population)):
            # run optimization
            cur_individual = population[pop_i]
            all_pop_i = 0
            if all_population is not None:
                while all_population[all_pop_i] != cur_individual:
                    all_pop_i += 1
            self.load_state(model, initial_state)
            model.set_kernel_function(cur_individual)
            if verbose:
                print(model.get_kernel_function())

            nested_log = None
            best_state['fun_calls'] += 1
            try:
                if self.nested_fit_method is not None:
                    nested_log = self.nested_fit_method.fit(
                        model,
                        folds,
                        budget=max_fun_call - best_state['fun_calls'],
                        verbose=verbose
                    )
                    best_state['fun_calls'] += nested_log['fun_calls']
                fitness = self.obj_fun(
                    model=model,
                    folds=folds,
                    grad_needed=False
                )
                if not hasattr(fitness, "__len__"):
                    fitness = [fitness]
                if verbose:
                    print(fitness)
            except (AssertionError, np.linalg.linalg.LinAlgError) as ex:
                fitness = [KernelFittingMethod.MAX_SIGMA]
                if verbose:
                    print(ex)

            individual = model.get_kernel_function()
            individual.fitness.setValues(fitness)
            individual.nested = nested_log
            individual.log.evals += 1
            population[pop_i] = individual
            if all_population is not None:
                all_population[all_pop_i] = individual

    def model_to_state(self, model):
        """

        :param model:
        :type model:
        :return:
        :rtype:
        """
        start = time.time()

        best_state = {
            'name': self.__class__.__name__,
            'fun_calls': 0,
            'value': np.inf,
            'time': start,
            'creation': 0,
            'evals': 0,
            'id': 0,
            'origin': 0,
            'hp_count': 0,
            'prim_count': 0
        }
        best_state.update(self.save_state(model))
        initial_state = self.save_state(model)
        return best_state, initial_state

    def best_to_model(self, model, best_state, best):
        """

        :param model:
        :type model:
        :param best_state:
        :type best_state:
        :param best:
        :type best:
        :return:
        :rtype:
        """
        end = time.time()

        best_state['value'] = best.fitness.getValues()
        best_state['time'] = end - best_state['time']
        best_state['creation'] = best.log.creation
        best_state['evals'] = best.log.evals
        best_state['id'] = best.log.id
        best_state['origin'] = best.log.origin
        best_state['hp_count'] = best.log.hp_count
        best_state['prim_count'] = best.log.prim_count
        best_state['state'] = str(best)
        best_state['nested'] = best.nested

        self.load_state(model, best_state)

    def save_state(self, model):
        """

        :param model:
        :type model:
        :return:
        :rtype:
        """
        if not isinstance(model.get_kernel_function(), deap.creator.Individual):
            raise Exception("Kernel is not Individual")

        state = self.nested_save_state(model)
        state['state'] = str(model.get_kernel_function())

        return state

    def load_state(self, model, state):
        """

        :param model:
        :type model:
        :param state:
        :type state:
        :return:
        :rtype:
        """
        kernel = state['state']
        if not isinstance(kernel, str):
            raise Exception("Kernel is not str")

        kernel = self.toolbox.individual_from_string(kernel)
        model.set_kernel_function(kernel)
        self.nested_load_state(model, state)

    def get_max_eval(self, budget):
        """

        :param budget:
        :type budget:
        :return:
        :rtype:
        """
        if budget is None or self.max_fun_call < budget:
            max_fun_call = self.max_fun_call
        else:
            max_fun_call = budget

        max_eval = max_fun_call
        if self.nested_fit_method is not None:
            max_eval = int(
                max_fun_call / (self.nested_fit_method.max_fun_call + 1)
            )

        return max_eval, max_fun_call

    def get_random_kernel(self):
        """

        :return:
        :rtype:
        """
        return self.toolbox.random_individual()

    def get_kernel_from_str(self, kernel_str):
        """

        :param kernel_str:
        :type kernel_str:
        :return:
        :rtype:
        """
        return self.toolbox.individual_from_string(kernel_str)

    def fit(self, model, folds, budget=None, verbose=False):
        """

        :param model:
        :type model:
        :param folds:
        :type folds:
        :param budget:
        :type budget:
        :param verbose:
        :type verbose:
        :return:
        :rtype:
        """

        raise NotImplementedError("Not Implemented. This is an interface.")



# -*- coding: utf-8 -*-
#
#    Copyright 2020 Ibai Roman
#
#    This file is part of EvoCov.
#
#    EvoCov is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    EvoCov is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with EvoCov. If not, see <http://www.gnu.org/licenses/>.

import deap.tools


def add_selection(toolbox, selection_method):
    """

    :param toolbox:
    :type toolbox:
    :param selection_method:
    :type selection_method:
    :return:
    :rtype:
    """
    if selection_method == 'best':
        toolbox.register("select", deap.tools.selBest)
    elif selection_method == 'NSGA2':
        toolbox.register("select", deap.tools.selNSGA2)
    else:
        raise Exception("This selection method is not available")

    toolbox.register("get_nondominated", get_nondominated)

def get_nondominated(population):
    """

    :param population:
    :type population:
    :return:
    :rtype:
    """
    return deap.tools.sortNondominated(
        population,
        len(population),
        first_front_only=True
    )[0]
